package com.abssqr.plat.biz.shared.dispatcher;

import com.abssqr.plat.biz.shared.scheduler.BaseQuartzJob;
import com.abssqr.plat.common.dal.mysql.auto.resultmap.HostHeartInfo;
import com.abssqr.plat.common.dal.mysql.auto.resultmap.HostTaskGroupCnt;
import com.abssqr.plat.common.facade.enums.TaskStatusEnum;
import com.abssqr.plat.common.model.domain.job.JobControl;
import com.abssqr.plat.common.model.domain.task.TaskGroup;
import com.abssqr.plat.common.model.repo.job.JobControlRepository;
import com.abssqr.plat.common.model.repo.task.TaskGroupRepo;
import com.abssqr.plat.core.service.host.HostManager;
import com.general.system.common.util.LogUtil;
import com.general.system.common.util.VarChecker;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.quartz.PersistJobDataAfterExecution;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.transaction.support.TransactionTemplate;

import java.util.*;
import java.util.stream.Collectors;

@PersistJobDataAfterExecution
public abstract class DispatcherSplitor<T> extends BaseQuartzJob {

    private final static Logger LOGGER             = LoggerFactory.getLogger(DispatcherSplitor.class);
    //心跳时间间隔 单位（秒）
    public static final  int    HEAT_BEAT_INTERVAL = 10;

    //处理时效的心跳间隔
    public static final int DISABLE_HEAT_BEAT_AMT = 5;

    //处理的心跳时间间隔
    public static final int RE_INIT_HEAT_BEAT_AMT = 6;

    public static final int HOST_GROUP_CNT = 2;

    @Autowired
    TaskGroupRepo        taskGroupRepo;
    @Autowired
    HostManager          hostManager;
    @Autowired
    JobControlRepository jobControlRepository;
    @Autowired
    private TransactionTemplate transactionTemplate;

    @Override
    protected void doExecute() {
        // LogUtil.info(LOGGER, "[{0}-DispatcherSplitor]-start", getJobName());

        transactionTemplate.execute(status -> {
            //每个Host拥有几个任务组
            Map<String, Integer> hostGroupCntMap = new HashMap<>();
            //加锁防止其他人直接调用这个方法
            JobControl jobControl = jobControlRepository.lockByJobName(getJobName());
            if (jobControl == null) {
                return status;
            }
            List<HostHeartInfo> allHost = hostManager.getAllHost();
            VarChecker.checkNotEmpty(allHost, "心跳任务曾未启动");

            //获取所有未完成的任务
            List<TaskGroup> unFinishedTaskGroups = taskGroupRepo.getUnFinishedTaskGroups(getJobName());

            Map<String, List<HostHeartInfo>> hostMap = allHost.stream().collect(
                Collectors.groupingBy(HostHeartInfo::getHostName));

            //如果没有任何任务
            if (CollectionUtils.isEmpty(unFinishedTaskGroups)) {
                //LogUtil.info(LOGGER, "[{0}--没有任务组，正在获取任务组]", getJobName());
                unFinishedTaskGroups = this.split();
                //LogUtil.info(LOGGER, "[{0}--获取任务成功，获取了{1}个任务组]", getJobName(), CollectionUtils.isNotEmpty
                // (unFinishedTaskGroups) ? unFinishedTaskGroups.size() : 0);
                if (CollectionUtils.isNotEmpty(unFinishedTaskGroups)) {
                    initTaskGroup(unFinishedTaskGroups);
                    taskGroupRepo.saveTaskGroups(unFinishedTaskGroups);
                }
                unFinishedTaskGroups = taskGroupRepo.getUnFinishedTaskGroups(getJobName());
            }
            if (CollectionUtils.isEmpty(unFinishedTaskGroups)) {
                //LogUtil.info(LOGGER, "[{0}-DispatcherSplitor]-end", getJobName());
                return status;
            }
            List<TaskGroup> freeTaskGroup = new ArrayList<>();
            for (TaskGroup taskGroup : unFinishedTaskGroups) {
                //已分配
                if (StringUtils.isNotBlank(taskGroup.getHostName())) {
                    HostHeartInfo hostHeartInfo = hostMap.get(taskGroup.getHostName()).stream().findFirst()
                        .get();
                    Integer hostGroupCnt = hostGroupCntMap.get(hostHeartInfo);
                    if (hostGroupCnt != null) {
                        hostGroupCntMap.put(hostHeartInfo.getHostName(), hostGroupCnt + 1);
                    } else {
                        hostGroupCntMap.put(hostHeartInfo.getHostName(), 1);
                    }
                    if (!isActive(hostHeartInfo)) {
                        //已处理过失效
                        if (taskGroup.getDisable()) {
                            if (canReInit(hostHeartInfo)) {
                                //标记当前任务为初始化状态
                                taskGroupRepo.markTaskGroupInit(taskGroup.getId());
                                LogUtil.info(LOGGER, "[{0}--机器{1}超过三个心跳未响应且任务组{2}的状态为disable，标记当前任务状态为初始化]",
                                    getJobName(), hostHeartInfo.getHostIp(),
                                    CollectionUtils.isNotEmpty(unFinishedTaskGroups) ? unFinishedTaskGroups.size() : 0);
                            }

                        } else {
                            //标记当前任务组状态为Disable=true
                            taskGroupRepo.markTaskGroupDisable(taskGroup.getId());
                            LogUtil.info(LOGGER, "[{0}--机器{1}超过三个心跳未响应且任务组{2}的状态为enable，标记当前任务状态为disable]",
                                getJobName(), hostHeartInfo.getHostIp(),
                                CollectionUtils.isNotEmpty(unFinishedTaskGroups) ? unFinishedTaskGroups.size() : 0);

                        }
                    } else {
                        if (taskGroup.getDisable()) {
                            taskGroupRepo.markTaskGroupEnable(taskGroup.getId());
                            LogUtil.info(LOGGER, "[{0}--机器{1}在第三个心跳在任务组{2}的状态为enable，将任务重置为enable", getJobName(),
                                hostHeartInfo.getHostIp(),
                                CollectionUtils.isNotEmpty(unFinishedTaskGroups) ? unFinishedTaskGroups.size() : 0);
                        }
                    }
                    continue;
                }
                freeTaskGroup.add(taskGroup);
            }
            List<HostTaskGroupCnt> hostTaskGroupCnt = taskGroupRepo.getHostTaskGroupCnt(getJobName());
            //获取所有的host拥有的任务组数
            Map<String, Integer> hostTaskGrooCnt = getHostTaskGrooCnt(allHost, hostTaskGroupCnt);
            hostTaskGrooCnt.forEach((k, v) -> {
                //如果小于两个则去分配任务组
                if (v < HOST_GROUP_CNT) {
                    List<Long> idUseCnt = getIdUseCnt(freeTaskGroup, HOST_GROUP_CNT - v);
                    idUseCnt.forEach(taskGroupId -> {
                        taskGroupRepo.distributeTaskGroup(taskGroupId, k);
                    });
                }
            });

            LogUtil.info(LOGGER, "[{0}-DispatcherSplitor]-end", getJobName());
            return status;
        });

    }

    private Map<String, Integer> getHostTaskGrooCnt(List<HostHeartInfo> allHost,
                                                    List<HostTaskGroupCnt> hostTaskGroupCnts) {
        Map<String, Integer> hostGroupCnt = new HashMap<>();
        Map<String, Integer> existsHostGroupCnt = hostTaskGroupCnts.stream()
            .collect(Collectors.toMap(HostTaskGroupCnt::getHostName, HostTaskGroupCnt::getGroupCnt));
        allHost.forEach(v -> {
            if (isActive(v)) {
                Integer groupCnt = existsHostGroupCnt.get(v.getHostName());
                if (groupCnt == null) {
                    groupCnt = 0;
                }
                hostGroupCnt.put(v.getHostName(), groupCnt);
            }
        });
        return hostGroupCnt;
    }

    private List<Long> getIdUseCnt(List<TaskGroup> freeTaskGroups, int cnt) {
        List<Long> idList = new ArrayList<>();

        int startIndex = 0;
        int endIndex = freeTaskGroups.size();
        List<Long> freeTaskIds;
        if (freeTaskGroups.size() > cnt) {
            endIndex = cnt;
        }
        List<TaskGroup> freeTaskGroupCnt = freeTaskGroups.subList(startIndex, endIndex);
        freeTaskIds = freeTaskGroupCnt.stream().map(v -> v.getId())
            .collect(Collectors.toList());
        idList.addAll(freeTaskIds);
        freeTaskGroups.removeAll(freeTaskGroupCnt);
        return idList;
    }

    public void initTaskGroup(List<TaskGroup> taskGroups) {
        taskGroups.forEach(v -> {
            v.setDisable(false);
            v.setStatus(TaskStatusEnum.INIT);
        });
    }

    public abstract List<TaskGroup> split();

    public abstract String getTaskJobName();

    public List<TaskGroup> getTask() {
        return taskGroupRepo.getTaskGroup(getTaskJobName(), hostManager.getHostInfo().getHostName());
    }

    public void finish(Long id) {
        taskGroupRepo.markTaskGroupFinish(id);

    }

    private boolean isActive(HostHeartInfo hostHeartInfo) {
        return HEAT_BEAT_INTERVAL * DISABLE_HEAT_BEAT_AMT > hostHeartInfo.getUnHeartSecond();
    }

    private boolean canReInit(HostHeartInfo hostHeartInfo) {
        return HEAT_BEAT_INTERVAL * RE_INIT_HEAT_BEAT_AMT > hostHeartInfo.getUnHeartSecond();
    }
}
