/**
 * abssqr.com Inc.
 * Copyright (c) 2017-2020 All Rights Reserved.
 */

package com.abssqr.plat.biz.shared.handler;

import com.abssqr.plat.biz.shared.dispatcher.buyBackMatch.LoanAstBuyBackMatchStarter;
import com.abssqr.plat.common.dal.mysql.auto.dataobject.LoanBakDO;
import com.abssqr.plat.common.dal.mysql.auto.paging.LoanBakListPage;
import com.abssqr.plat.common.dal.mysql.auto.resultmap.AstPoolItem;
import com.abssqr.plat.common.facade.enums.*;
import com.abssqr.plat.common.facade.model.org.OrganizationEntity;
import com.abssqr.plat.common.facade.model.plan.Plan;
import com.abssqr.plat.common.facade.model.plan.PlanBase;
import com.abssqr.plat.common.facade.model.rule.AstPriceRule;
import com.abssqr.plat.common.facade.model.rule.AstScreenRule;
import com.abssqr.plat.common.facade.model.rule.TradePriceRule;
import com.abssqr.plat.common.model.domain.ast.*;
import com.abssqr.plat.common.model.domain.contract.LoanContract;
import com.abssqr.plat.common.model.domain.job.JobControl;
import com.abssqr.plat.common.model.domain.rule.LoanAstPriceCalcParam;
import com.abssqr.plat.common.model.domain.rule.LoanAstPriceResult;
import com.abssqr.plat.common.model.domain.task.TrCtrTask;
import com.abssqr.plat.common.model.domain.task.TrTask;
import com.abssqr.plat.common.model.enums.IdTypeEnum;
import com.abssqr.plat.common.model.repo.ast.LoanRepository;
import com.abssqr.plat.common.model.repo.ast.PackRepository;
import com.abssqr.plat.common.model.repo.contract.LoanContractRepository;
import com.abssqr.plat.common.model.repo.job.JobControlRepository;
import com.abssqr.plat.common.model.repo.task.AstSyncTaskCtrRepo;
import com.abssqr.plat.common.model.repo.tr.TrCtrTaskRepo;
import com.abssqr.plat.common.model.repo.tr.TrTaskRepo;
import com.abssqr.plat.common.model.seq.SequenceUtil;
import com.abssqr.plat.common.model.utils.TextUtil;
import com.abssqr.plat.common.util.profiler.PerformTrace;
import com.abssqr.plat.core.service.accounting.OrgAcctDayComponent;
import com.abssqr.plat.core.service.ast.LoanAstAssembler;
import com.abssqr.plat.core.service.ast.LoanAstBuyBackManager;
import com.abssqr.plat.core.service.plan.PlanManager;
import com.abssqr.plat.core.service.rule.AstPriceRuleComponent;
import com.abssqr.plat.core.service.rule.AstScreenRuleComponent;
import com.general.system.common.util.*;
import com.general.system.tool.util.ToolUtil;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import org.apache.commons.collections.CollectionUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import org.springframework.transaction.support.TransactionTemplate;

import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.function.Consumer;

import static com.abssqr.plat.common.model.enums.IdTypeEnum.ASTSCREENINGRULE;

/**
 * 回购
 *
 * @author xiachenxiang
 * @version com.abssqr.plat.biz.shared.handler: LoanAstBuyBackMatchHandler.java, v 0.1 2020-02-10 1:50 PM xiachenxiang Exp $
 */
@Component
public class LoanAstBuyBackMatchHandler {
    private final Logger                 LOGGER = LoggerFactory.getLogger(getClass());
    @Autowired
    private       TrTaskRepo             trTaskRepo;
    @Autowired
    private       PackRepository         packRepository;
    @Autowired
    private       LoanAstBuyBackManager  loanAstBuyBackManager;
    @Autowired
    private       AstSyncTaskCtrRepo     astSyncTaskCtrRepo;
    @Autowired
    private       LoanRepository         loanRepository;
    @Autowired
    private       LoanContractRepository loanContractRepository;
    @Autowired
    private       TrCtrTaskRepo          trCtrTaskRepo;
    @Autowired
    private       AstScreenRuleComponent astScreenRuleComponent;
    @Autowired
    private       TransactionTemplate    transactionTemplate;
    @Autowired
    private       AstPriceRuleComponent  astPriceRuleComponent;
    @Autowired
    private       LoanAstAssembler       loanAstAssembler;
    @Autowired
    private       JobControlRepository   jobControlRepository;
    @Autowired
    private       OrgAcctDayComponent    orgAcctDayComponent;
    @Autowired
    private       PlanManager            planManager;


    private final int limit          = 2000;
    private       int batchInsertCnt = 500;

    @PerformTrace(name = "LoanAstBuyBackMatchHandler")
    public void execute(String packCode, Consumer<TaskStatusEnum> consumer) {
        // 1.获取资产包
        LoanAstBuyBackPack existPack = loanAstBuyBackManager.getByPackCode(packCode);
        VarChecker.checkNotNull(existPack, "[LoanAstBuyBackMatchHandler]-回购筛选启动失败,没有对应的Pack");
        // 2.判断状态
        VarChecker.checkEquals(existPack.getPackStatus(), AstPackStatusEnum.INIT, "[LoanAstBuyBackMatchHandler]-回购筛选启动失败,Pack[{0}]的状态不为init", existPack.getPackCode());
        // 3.清理历史TrTask
        Date curDate = orgAcctDayComponent.getTrfAcctDay().getStandardDate();
        trTaskRepo.clearHisTrTask(existPack.getBackOutCode(), curDate, AstPackTradeTypeEnum.BUY_BACK_BIZ);

        Map<String, TrCtrTask> trCtrTaskMap = Maps.newHashMap();
        String batchNo = SequenceUtil.genId(ASTSCREENINGRULE);
        LogUtil.info(LOGGER, "开始筛选[{0}]受让计划，批次号=[{1}]", existPack.getBackOutCode(), batchNo);

        // 校验是否完成模型同步
        this.checkAstModFinished(existPack, curDate);


        // 4.分页遍历承接池机构资产
        LoanBakListPage page = new LoanBakListPage();
        page.setAccountDate(curDate);
        page.setOwnOrg(existPack.getBackOutCode());
        page.setStatuList(Lists.newArrayList(LoanStatusEnum.NORMAL.getCode(), LoanStatusEnum.OVERDUE.getCode()));
        page.setLimit(limit);
        page.setCurrPageNo(1);

        do {
            if (!this.canRun()) {
                LogUtil.info(LOGGER, "[LoanAstBuyBackMatchHandler] - 检测到[{0}]任务已停止，跳出筛选", LoanAstBuyBackMatchStarter.JOB_NAME);
                break;
            }
            LogUtil.info(LOGGER, "[LoanAstBuyBackMatchHandler]-[{0}]开始筛选[{0}]受让计划第[{1}]页", existPack.getBackOutCode(), page.getCurrentPage());
            this.astMatch(page, existPack, trCtrTaskMap, curDate, batchNo);
            page.setCurrPageNo(page.getCurrPageNo() + 1);

        } while (page.hasNextPage());

        LogUtil.info(LOGGER, "[LoanAstBuyBackMatchHandler] - 筛选结束");

        // 5 更新转让数据状态
        transactionTemplate.execute(status -> {
            packRepository.updatePackStatus(Lists.newArrayList(existPack), AstPackStatusEnum.SCEEN_FIS, AstPackStatusEnum.INIT);
            trTaskRepo.updateStatus(batchNo, TaskStatusEnum.INIT, TaskStatusEnum.IGNORE);
            //插入转让控制表，可以开始转让
            if (ToolUtil.isNotEmpty(trCtrTaskMap)) {
                trCtrTaskRepo.insertBatch(Lists.newArrayList(trCtrTaskMap.values()));
            }
            consumer.accept(TaskStatusEnum.SUCCESS);
            return status;
        });

    }

    /**
     * 判断当前机构是否完成模型同步
     *
     * @param pack    资产包
     * @param curDate 会计日
     */
    private void checkAstModFinished(LoanAstBuyBackPack pack, Date curDate) {
        // 判断当日模型同步情况
        List<String> orgCodes = astSyncTaskCtrRepo.getLastStatusFinishedOrgCodes(curDate, SyncTaskTypeEnum.AST_MOD);
        VarChecker.checkNotEmpty(orgCodes, "[LoanAstBuyBackMatchHandler]-回购筛选启动失败,当日没有机构完成模型同步");

        // 判断当前转让会计日是否完成模型转换，未完成直接退出
        PlanBase buybackPlan = planManager.getPlanFullByNo(pack.getPlanNo());
        String OrgCode = null;
        // 放款池回购受让池
        if (buybackPlan.getPlanType() == PlanTypeEnum.SPV && buybackPlan instanceof OrganizationEntity) {
            OrgCode = ((OrganizationEntity) buybackPlan).getManagerOrgCode();
        } else { // 受让池回购受让池
            // 判断两受让池对应的放款池是否一致
            PlanBase outplan = planManager.getPlanFullByNo(pack.getBackOutCode());
            VarChecker.checkArgument(outplan.getPlanType() == PlanTypeEnum.PLAN && outplan instanceof Plan, "[LoanAstBuyBackMatchHandler]-回购筛选启动失败,不支持的回购类型受让池回购放款池");
            Optional<AstScreenRule> outTrfScreenRuleOptional = ((Plan) outplan).getPlanTransferRule().getAstScreenRuleList().stream().filter(item -> item.getScreenType() == ScreenTypeEnum.AST_TRF).findFirst();
            VarChecker.checkArgument(outTrfScreenRuleOptional.isPresent(), "[LoanAstBuyBackMatchHandler]-回购筛选启动失败,转让筛选规则不存在");
            AstScreenRule astScreenRule = outTrfScreenRuleOptional.get();
            VarChecker.checkEquals(astScreenRule.getLenderOrgCode(), pack.getLoanProduct().getLenderOrgCode(), "[LoanAstBuyBackMatchHandler]-回购筛选启动失败,两个受让计划不属于同一个放款池");
            OrgCode = astScreenRule.getManagerOrgCode();

        }
        VarChecker.checkArgument(orgCodes.contains(OrgCode), "[LoanAstBuyBackMatchHandler]-回购筛选启动失败,机构[{0}]未完成模型同步", ((OrganizationEntity) buybackPlan).getManagerOrgCode());
    }

    /**
     * 检测是否继续执行筛选
     *
     * @return
     */
    public boolean canRun() {
        JobControl jobControl = jobControlRepository.selectByJobName(LoanAstBuyBackMatchStarter.JOB_NAME);
        return jobControl.getIsEff().isYes();
    }

    private void astMatch(LoanBakListPage page, LoanAstBuyBackPack pack, Map<String, TrCtrTask> trCtrTaskMap, Date curDate, String batchNo) {
        page = loanRepository.getPageByOwnOrg(page);
        List<LoanBakDO> datas = page.getDatas();
        List<TrTask> trTasks = Lists.newArrayList();

        for (LoanBakDO loanBakDO : datas) {
            // 5.1.根据筛选规则进行匹配
            TrTask trTask = null;

            boolean isMatch = astScreenRuleComponent.astBuybackScreen(curDate, pack, () -> buildAstPoolItem(curDate, loanBakDO));
            if (isMatch) {

                // 5.2.匹配成功资产按照定价规则进行定价
                LoanAstPriceResult result = this.makePrice(loanBakDO.getAstNo(), pack);
                if (result == null) {
                    LogUtil.info(LOGGER, "[LoanAstBuyBackMatchHandler] - 计划[{0}]下机构资产[{1}]为0,跳过", loanBakDO.getOwnOrg(), loanBakDO.getAstNo());
                    continue;
                }
                TrCtrTask trCtrTask = trCtrTaskMap.get(pack.getPackCode());
                if (trCtrTask == null) {
                    trCtrTask = new TrCtrTask();
                    trCtrTaskMap.put(pack.getPackCode(), trCtrTask);
                    trCtrTask.setTfrDate(curDate);
                    trCtrTask.setBatchNo(batchNo);
                    trCtrTask.setPackCode(pack.getPackCode());
                    trCtrTask.setPackType(pack.getPackType());
                    trCtrTask.setTaskStatus(TaskStatusEnum.INIT);
                    trCtrTask.setManagerOrgCode(pack.getManagerOrgCode());
                    trCtrTask.setPackType(AstPackTradeTypeEnum.BUY_BACK_BIZ);
                }
                if (MoneyUtil.compare(trCtrTask.getPreTrAmt().add(result.getPrice().getSum()), pack.getAvalAmt()) <= 0) {
                    trCtrTask.addPreTrAmt(result.getPrice().getSum());
                    trTask = dealTrTask(loanBakDO, batchNo, curDate, pack, result);
                }
            }
            if (trTask != null) {
                VarChecker.checkArgument(trTask.getTfrPrinBal().isGreaterThanZero(),
                        "Pack筛选到资产-待转金额为0,trTask[{0}]", trTask);
                trTasks.add(trTask);
                LogUtil.info(LOGGER, "[{0}]Pack筛选到资产，astNo=[{1}],batchNo=[{2}]", trTask.getTfrPack(),
                        trTask.getAstNo(), batchNo);
                trySaveTrTask(trTasks, batchInsertCnt);
            }
        }
        trySaveTrTask(trTasks, 0);

    }

    /**
     * 计算定价
     *
     * @param astNo 资产编号
     * @param pack  回购资产包
     * @return
     */
    private LoanAstPriceResult makePrice(String astNo, LoanAstBuyBackPack pack) {
        Loan loan = loanRepository.getByAstNo(astNo);
        loan = loanAstAssembler.assembleLoan(loan, orgAcctDayComponent.getTrfAcctDay().getStandardDate());
        LoanAstGroup loanAstGroup = loan.filterLoanAstByOrg(pack.getBackOutCode(), loan.getTermCnt());
        LoanAstPriceResult result = null;
        for (LoanAst loanAst : loanAstGroup.getLoanAsts()) {
            LoanAstPriceCalcParam param = new LoanAstPriceCalcParam();
            param.setTransIn(pack.getPlanNo());
            param.setTransOut(pack.getBackOutCode());
            param.setLoanAst(loanAst);
            param.setTradePriceRules(pack.getAstPriceRuleList());
            param.setAstTradeType(AstPackTradeTypeEnum.BUY_BACK_BIZ);
            // 资产五级分类
            param.setCondition(loan.getLoanRisk().getRiskClass().getCode());
            if (null == result) {
                result = astPriceRuleComponent.execute(param);
            } else {
                result.merge(astPriceRuleComponent.execute(param));
            }
        }

        return result;

    }


    /**
     * 尝试保存
     *
     * @param trTasks
     * @param minSize
     */
    private void trySaveTrTask(List<TrTask> trTasks, int minSize) {
        if (CollectionUtils.isNotEmpty(trTasks) && trTasks.size() >= minSize) {
            LogUtil.info(LOGGER, "LoanAstBuyBackMatchHandler 回购TrTask入库size={0},minSize={1}", trTasks.size(), minSize);
            trTaskRepo.insertBatch(trTasks);
            trTasks.clear();
        }
    }

    private TrTask dealTrTask(LoanBakDO loanBakDO, String batchNo, Date acctDate,
                              LoanAstBuyBackPack pack, LoanAstPriceResult result) {
        TrTask task = new TrTask();
        task.setExecTimes(0L);
        task.setTfrDt(acctDate);
        task.setBatchNo(batchNo);
        task.setNextExecTime(SystemDateUtil.getSystemDate());
        task.setAstNo(loanBakDO.getAstNo());
        task.setTfrPack(pack.getPackCode());
        task.setTermNo(loanBakDO.getTermCnt());
        task.setAstDueDt(loanBakDO.getEndDate());
        task.setTfrPrinBal(result.getBal().getAllPrin());
        task.setTaskStatus(TaskStatusEnum.IGNORE);
        task.setTfrOut(loanBakDO.getOwnOrg());
        task.setSeqNo(SequenceUtil.genId(IdTypeEnum.SEQ));
        task.setPackType(AstPackTradeTypeEnum.BUY_BACK_BIZ);
        TradePriceRule tradePriceRule = result.getTradePriceRule();
        VarChecker.checkArgument(tradePriceRule instanceof AstPriceRule, "[{0}]回购筛选交易定价类型有误", pack.getBackOutCode());
        AstPriceRule astPriceRule = (AstPriceRule) tradePriceRule;
        task.setTfrIntrBal(result.getBal().getAllIntr());
        task.setTfrPenalFeeBal(result.getBal().getAllIntAndFee().subtract(result.getBal().getAllIntr()));
        task.setRiskCl(astPriceRule.getRiskCl());
        task.setIntrRate(astPriceRule.getIntrRate());
        task.setPrinRate(astPriceRule.getPrinRate());
        task.setPenalFeeRate(astPriceRule.getPenalFeeRate());

        task.setTfrAmt(result.getPrice().getSum());

        return task;
    }

    /**
     * 构建资产条目
     *
     * @param curDate
     * @param loanBakDO
     * @return
     */
    private AstPoolItem buildAstPoolItem(Date curDate, LoanBakDO loanBakDO) {
        LoanContract contract = loanContractRepository.getByAstNo(loanBakDO.getAstNo());
        // 需要
        AstPoolItem astPoolItem = new AstPoolItem();
        astPoolItem.setStatus(loanBakDO.getStatus());
        astPoolItem.setAstNo(loanBakDO.getAstNo());
        astPoolItem.setProdNo(loanBakDO.getProdNo());
        astPoolItem.setTermNo(loanBakDO.getTermCnt());
        astPoolItem.setStartDate(loanBakDO.getStartDate());
        astPoolItem.setEndDate(loanBakDO.getEndDate());
        astPoolItem.setRemainDays((int) DateTimeUtil.calcDayDiff(astPoolItem.getEndDate(), curDate));
        astPoolItem.setPrinBal(loanBakDO.getPrinBal());
        //astPoolItem.setIntrBal(loan.getBal().getAllIntr());
        //astPoolItem.setPrinPenalBal(loan.getBal().getByFieldType(DebtMoneyFieldEnum.PRIN_ODUE_PNY));
        //astPoolItem.setIntrPenalBal(loan.getBal().getByFieldType(DebtMoneyFieldEnum.INTR_ODUE_PNY));
        astPoolItem.setManagerOrgCode(loanBakDO.getManagerOrgCode());
        astPoolItem.setLenderOrgCode(loanBakDO.getLenderOrgCode());
        astPoolItem.setTermCnt(loanBakDO.getTermCnt());
        astPoolItem.setRiskCl(loanBakDO.getRiskCl());
        Map<String, String> context = Maps.newHashMap();
        context.put(AstConditions.CERT_NO, contract.getLoanCustInfo().getCertNo());
        context.put(AstConditions.CERT_TYPE, contract.getLoanCustInfo().getCertType().getCode());
        context.put(AstConditions.INTER_RATE, contract.getLoanProperty().getRate().toString());
        context.put(AstConditions.RATE_TYPE, contract.getLoanProperty().getRateType().getCode());
        astPoolItem.setContext(TextUtil.convertMap2Txt(context));
        return astPoolItem;
    }
}
