/**
 * abssqr.com Inc.
 * Copyright (c) 2017-2019 All Rights Reserved.
 */
package com.abssqr.plat.biz.shared.handler;

import com.abssqr.plat.common.facade.enums.AstPackStatusEnum;
import com.abssqr.plat.common.facade.enums.TrCycleEnum;
import com.abssqr.plat.common.facade.enums.TransCodeEnum;
import com.abssqr.plat.common.facade.model.plan.PlanAst;
import com.abssqr.plat.common.facade.model.plan.PlanBase;
import com.abssqr.plat.common.facade.model.plan.PlanTransferRule;
import com.abssqr.plat.common.facade.model.rule.AstScreenRule;
import com.abssqr.plat.common.model.domain.ast.Pack;
import com.abssqr.plat.common.model.repo.clearing.PlanClearingLogRepo;
import com.abssqr.plat.common.model.repo.plan.PlanAstRepository;
import com.abssqr.plat.common.model.seq.SequenceUtil;
import com.abssqr.plat.core.service.ast.AstPackManager;
import com.abssqr.plat.core.service.plan.PlanInvestStructureManager;
import com.abssqr.plat.core.service.util.RptBuilderUtil;
import com.general.system.common.model.Money;
import com.general.system.common.util.DateTimeUtil;
import com.general.system.common.util.LogUtil;
import com.general.system.common.util.MessageUtil;
import com.general.system.common.util.VarChecker;
import org.apache.commons.collections.CollectionUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.util.Date;

/**
 * 计划资产创建资产包
 * @author zhenxuan.luo
 * @version com.abssqr.plat.biz.shared.handler: PlanAstReqHandler.java, v 0.1 2019-06-19 01:34 zhenxuan.luo Exp $
 */
@Component
public class PlanAstReqHandler {

    private final static Logger LOGGER = LoggerFactory.getLogger(PlanAstReqHandler.class);

    @Autowired
    private PlanAstRepository          planAstRepository;
    @Autowired
    private AstPackManager             astPackManager;
    @Autowired
    private PlanClearingLogRepo        planClearingLogRepo;
    @Autowired
    private PlanInvestStructureManager planInvestStructureManager;

    public Pack execute(PlanBase planBase, PlanTransferRule transferRule, Date accountDate) {

        Date tfrClearingDate = planClearingLogRepo.getMinUnStateAccountDate(planBase.getPlanNo(), null);
        if (tfrClearingDate != null) {
            LogUtil.info(LOGGER, "计划[{0}]创建资产包, 在[{1}]有未完成的转让任务待清算悬挂", planBase.getPlanNo(), tfrClearingDate);
            return null;
        }

        VarChecker.checkNotNull(transferRule, "计划[{0}]创建资产包找不到对应转让规则", planBase.getPlanNo());

        PlanAst planAst = planAstRepository.getByPlanNo(planBase.getPlanNo());

        Pack pack = new Pack();
        pack.setPlanNo(planBase.getPlanNo());
        pack.setPackCode(SequenceUtil.genSeqNo(TransCodeEnum.OTHER.getCode()));
        pack.setPackName(MessageUtil.formatMsg("{0}-{1}资产包",
                planBase.getPlanName(), DateTimeUtil.formatYMD(accountDate)));
        pack.setTransferRule(transferRule);
        pack.setPackStatus(AstPackStatusEnum.INIT);
        pack.setScenDate(accountDate);
        pack.setMinEndDate(null);
        pack.setMaxEndDate(planBase.getEndDate());
        pack.setCalcIntrType(planBase.getCalcIntrType());

        if (CollectionUtils.isNotEmpty(transferRule.getAstScreenRuleList())) {
            AstScreenRule astScreenRule = transferRule.getAstScreenRuleList().stream().findFirst().get();
            pack.setManagerOrgCode(astScreenRule.getManagerOrgCode());
        }

        boolean isNeedPack;
        if (transferRule.getTrCycle() != TrCycleEnum.STATIC) {
            isNeedPack = buildPackForCycle(transferRule, planAst, pack);
        } else {
            isNeedPack = buildPackForStatic(planBase, planAst, pack);
        }
        if (isNeedPack) {
            return pack;
        }

        return null;
    }

    private boolean buildPackForStatic(PlanBase planBase, PlanAst planAst, final Pack pack) {
        // 填充计划投资结构
        planBase.setInvestStructureList(planInvestStructureManager.selectByPlanNo(planBase.getPlanNo()));
        // 获取计划投资结构 投入资本金额总额
        Money avalAmt = RptBuilderUtil.getInvestedAmt(planBase).subtract(planAst.getAccGrant());

        if (avalAmt.isGreaterThanZero()) {
            pack.setAvalAmt(avalAmt);
            return true;
        } else {
            return false;
        }
    }

    private boolean buildPackForCycle(PlanTransferRule transferRule, PlanAst planAst, final Pack pack) {

        // 1. 转让周期判断   数量会越来越大，OOM风险
        //List<Pack> packList = packRepository.getByPlanNo(transferRule.getPlanNo());
        Date lastDate = null;
        // 当天可以多次转让
        //packList = packList.stream().filter(item -> !DateTimeUtil.isSameDay(item.getScenDate(), pack.getScenDate()))
        //                   .collect(Collectors.toList());
        //if (CollectionUtils.isNotEmpty(packList)) {
        //
        //    Collections.sort(packList, Comparator.comparing(Pack::getScenDate));
        //    Pack lastPack = packList.get(packList.size() - 1);
        //    lastDate = lastPack.getScenDate();
        //}

        Money leastValAmt = new Money(0);
        if (!isNeedTransAccdFreq(transferRule, leastValAmt, pack.getScenDate())) {
            return false;
        }

        // 获取当前实际可用余额
        Money ableCash = planAst.containClearAbleBal();
        Money avalAmt = ableCash.subtract(transferRule.getResAmt());

        if (avalAmt.isGreaterThanZero() && avalAmt.greaterThan(leastValAmt)) {
            pack.setAvalAmt(avalAmt);
            Date minEndDate = null;
            Date maxEndDate = null;

            for (AstScreenRule astScreenRule : transferRule.getAstScreenRuleList()) {
                if (astScreenRule.getEarliestEndDate() != null) {
                    if (null == minEndDate
                        || DateTimeUtil.calcDayDiff(minEndDate, astScreenRule.getEarliestEndDate()) > 0) {
                        minEndDate = astScreenRule.getEarliestEndDate();
                    }
                }

                if (astScreenRule.getLatestEndDate() != null) {
                    if (null == maxEndDate
                        || DateTimeUtil.calcDayDiff(astScreenRule.getLatestEndDate(), maxEndDate) > 0) {
                        maxEndDate = astScreenRule.getLatestEndDate();
                    }
                }
            }

            pack.setMinEndDate(minEndDate);

            // 默认不能出头，填入值为计划到期日
            if (null != maxEndDate) {
                pack.setMaxEndDate(maxEndDate);
            }
            return true;
        } else {
            LogUtil.info(LOGGER, "Pack生成忽略PlanNo:{0},avalAmt:{1},leastValAmt:{2}", transferRule.getPlanNo(), avalAmt,
                leastValAmt);
            return false;
        }
    }

    /**
     *  目前一天资产同步一次，当资产每天同步多次时需要调整
     * @param planTransferRule
     * @param leastValAmt
     * @param curDate
     * @return
     */
    private boolean isNeedTransAccdFreq(PlanTransferRule planTransferRule, Money leastValAmt, Date curDate) {
        Pack lastPack = astPackManager.getLastDayPackByPlanNo(planTransferRule.getPlanNo(), curDate);
        // 首次转让
        if (lastPack == null) {
            return true;
        }

        long dayDiff = DateTimeUtil.calcDayDiff(curDate, lastPack.getScenDate());
        int compareDays = 0;
        switch (planTransferRule.getTrCycle()) {
            case D:
                compareDays = 1;
                break;
            case W:
                compareDays = 7;
                break;
            case M:
                compareDays = 30;
                break;
            case Q:
                compareDays = 90;
                break;
            case Y:
                compareDays = 360;
                break;
        }

        if (dayDiff >= compareDays) {
            return true;
        }

        return false;
    }
}