/**
 * abssqr.com Inc.
 * Copyright (c) 2017-2019 All Rights Reserved.
 */
package com.abssqr.plat.biz.shared.handler;

import com.abssqr.plat.common.facade.enums.OrgAstStatStatusEnum;
import com.abssqr.plat.common.facade.enums.SyncTaskTypeEnum;
import com.abssqr.plat.common.facade.model.plan.PlanAst;
import com.abssqr.plat.common.facade.model.plan.PlanBase;
import com.abssqr.plat.common.model.domain.plan.PlanAstStatTask;
import com.abssqr.plat.common.model.repo.ast.PlanAstStatTaskRepo;
import com.abssqr.plat.common.model.repo.clearing.PlanClearingLogRepo;
import com.abssqr.plat.common.model.repo.plan.PlanAstRepository;
import com.abssqr.plat.common.util.base.Day;
import com.abssqr.plat.core.service.plan.PlanManager;
import com.abssqr.plat.core.service.plan.PlanSyncTaskCtrManager;
import com.general.system.common.util.DateTimeUtil;
import com.general.system.common.util.LogUtil;
import com.general.system.common.util.SystemDateUtil;
import com.general.system.common.util.VarChecker;
import com.general.system.tool.util.ToolUtil;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.util.Date;

/**
 * 初始化每日统计清算流水任务
 *
 * @author zhenxuan.luo
 * @version com.abssqr.plat.biz.shared.handler: PlanAstStatTaskHandler.java, v 0.1 2019-07-14 16:33 zhenxuan.luo Exp $
 */
@Component
public class PlanAstStatTaskHandler {

    protected final Logger LOGGER = LoggerFactory.getLogger(PlanAstStatTaskHandler.class);

    @Autowired
    private PlanAstRepository      planAstRepository;
    @Autowired
    private PlanClearingLogRepo    planClearingLogRepo;
    @Autowired
    private PlanAstStatTaskRepo    planAstStatTaskRepo;
    @Autowired
    private PlanSyncTaskCtrManager planSyncTaskCtrManager;
    @Autowired
    private PlanManager            planManager;

    /**
     * 针对计划生成当日 统计任务
     *
     * @param planNo
     */
    public void handle(String planNo, Day trfDate) {
        // 查询锁水位表
        PlanAst planAst = planAstRepository.lockByPlanNo(planNo);
        VarChecker.checkNotNull(planAst, "计划[{0}]资产水位信息未初始化", planNo);
        // 查询计划下的资产统计任务
        PlanAstStatTask planAstStatTask = planAstStatTaskRepo.getPlanNo(planAst.getPlanNo());

        boolean isNew = false;
        // 1. 如果统计任务不存在，则新建
        if (null == planAstStatTask) {
            // 从未进行统计
            planAstStatTask = this.initplanAstStatTask(planAst);
            isNew = true;
        } else if (planAstStatTask.getStatus() != OrgAstStatStatusEnum.WAIT) {
            return;
        }

        // 2. 如果统计任务是新建的或者当前日期大于统计任务的会计日期，则重新捞取会计日期
        if (isNeedResetAccountDate(planAstStatTask.getAccountDate(), planAstStatTask.getPlanNo(), trfDate)) {
            Date minAcctDate = planClearingLogRepo.getMinUnStateAccountDate(planAstStatTask.getPlanNo(), null);

            if (null == minAcctDate) {
                // 没有待处理的流水，直接更新
                planAstStatTask.setAccountDate(trfDate.getStandardDate());
                planAst.setLastUpdateDate(planAstStatTask.getAccountDate());
                planAstRepository.update(planAst);
            } else {
                // 注意：有可能仍然是原来的会计日期，因为有部分流水没处理完
                if (DateTimeUtil.calcDayDiff(minAcctDate, planAstStatTask.getAccountDate()) <= 0) {
                    LogUtil.warn(LOGGER, "计划统计任务[{0}]初始化，重置日期后，仍然为当日，可能有部分任务阻塞", planAstStatTask);
                }
                planAstStatTask.setAccountDate(minAcctDate);
            }
        }

        // 重置任务状态为活跃
        resetTaskState(planAstStatTask);

        // 4. 若任务不存在，则生成任务。若任务存在，则将任务状态修订为执行中，等待dispatcher捞起。（如果已经在执行中则忽略）
        if (isNew) {
            planAstStatTaskRepo.createTask(planAstStatTask);
        } else {
            planAstStatTaskRepo.updateTask(planAstStatTask);
        }
    }

    private void resetTaskState(PlanAstStatTask planAstStatTask) {
        planAstStatTask.setStatus(OrgAstStatStatusEnum.ACTIVE);
        planAstStatTask.setLastExecTime(planAstStatTask.getNextExecTime());
        planAstStatTask.setNextExecTime(SystemDateUtil.getSystemDate());
        planAstStatTask.setLastHost(null);
        planAstStatTask.setExecTimes(0L);
    }

    private PlanAstStatTask initplanAstStatTask(PlanAst planAst) {
        PlanAstStatTask planAstStatTask = new PlanAstStatTask();
        planAstStatTask.setPlanNo(planAst.getPlanNo());
        planAstStatTask.setAccountDate(planAst.getLastUpdateDate());
        // 时间戳字段不做消费，只做登记便于排查问题
        planAstStatTask.setLastExecTms("");
        planAstStatTask.setMinId(0L);
        planAstStatTask.setMaxId(0L);
        planAstStatTask.setCnt(0L);

        return planAstStatTask;
    }

    /**
     * 判断是否重置日期
     *
     * @param taskAccountDate
     * @param planNo
     * @return
     */
    private boolean isNeedResetAccountDate(Date taskAccountDate, String planNo, Day trfDate) {

        Date standardTaskDate = DateTimeUtil.getStandardDate(taskAccountDate);
        // 当前转让会计日期 小于 任务日 不重置日期
        if (ToolUtil.isNotEmpty(standardTaskDate)
            && DateTimeUtil.calcDayDiff(trfDate.getStandardDate(), standardTaskDate) <= 0) {
            return false;
        }
        // 任务日跨天+10分钟 大于 转让会计日 不重置日期
        Date refDate = DateTimeUtil.addMinutes(DateTimeUtil.addDays(standardTaskDate, 1), 10);
        if (refDate.compareTo(trfDate.getDate()) > 0) {
            return false;
        }
        // 取转让会计日-1 作为机构会计日判断是否完成模型转换
        Date orgDate = trfDate.getYesterday().getStandardDate();
        PlanBase planBase = planManager.getPlanFullByNo(planNo);
        return planSyncTaskCtrManager.checkSyncFinishGtIsTrue(planBase, orgDate, SyncTaskTypeEnum.AST_MOD);

    }
}