/**
 * abssqr.com Inc.
 * Copyright (c) 2017-2019 All Rights Reserved.
 */
package com.abssqr.plat.biz.shared.scheduler.job.plan;

import com.abssqr.plat.biz.shared.handler.PlanAstReqHandler;
import com.abssqr.plat.biz.shared.scheduler.BaseQuartzJob;
import com.abssqr.plat.common.dal.mysql.auto.dao.PlanTranScreenRelDAO;
import com.abssqr.plat.common.facade.enums.AstPackStatusEnum;
import com.abssqr.plat.common.facade.enums.AstPackTradeTypeEnum;
import com.abssqr.plat.common.facade.enums.PlanTypeEnum;
import com.abssqr.plat.common.facade.enums.SyncTaskTypeEnum;
import com.abssqr.plat.common.facade.enums.TaskStatusEnum;
import com.abssqr.plat.common.facade.model.plan.PlanBase;
import com.abssqr.plat.common.facade.model.plan.PlanTransferRule;
import com.abssqr.plat.common.model.domain.ast.Pack;
import com.abssqr.plat.common.model.domain.task.TrCtrTask;
import com.abssqr.plat.common.model.repo.ast.PackRepository;
import com.abssqr.plat.common.model.repo.plan.PlanRepo;
import com.abssqr.plat.common.model.repo.task.AstSyncTaskCtrRepo;
import com.abssqr.plat.common.model.repo.tr.TrCtrTaskRepo;
import com.abssqr.plat.core.service.accounting.OrgAcctDayComponent;
import com.abssqr.plat.core.service.ast.AstPackManager;
import com.abssqr.plat.core.service.plan.PlanManager;
import com.abssqr.plat.core.service.plan.PlanTransferRuleManager;
import com.general.system.common.util.DateTimeUtil;
import com.general.system.common.util.DateTimeUtil.BetweenType;
import com.general.system.common.util.LogUtil;
import com.general.system.common.util.VarChecker;
import com.google.common.collect.Lists;
import org.apache.commons.collections.CollectionUtils;
import org.quartz.DisallowConcurrentExecution;
import org.quartz.PersistJobDataAfterExecution;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;
import org.springframework.transaction.TransactionStatus;
import org.springframework.transaction.support.TransactionCallbackWithoutResult;
import org.springframework.transaction.support.TransactionTemplate;

import java.util.Date;
import java.util.List;

/**
 * 计划资产需求统计任务
 *
 * @author zhenxuan.luo
 * @version com.abssqr.plat.biz.shared.scheduler.job.plan: PlanPackGenJob.java, v 0.1 2019-06-18 21:19 zhenxuan.luo
 * Exp $
 */
@PersistJobDataAfterExecution
@DisallowConcurrentExecution
@Component
public class PlanPackJob extends BaseQuartzJob {

    private final static Logger LOGGER = LoggerFactory.getLogger(PlanPackJob.class);

    @Autowired
    AstSyncTaskCtrRepo   astSyncTaskCtrRepo;
    @Autowired
    TransactionTemplate  transactionTemplate;
    @Autowired
    PlanAstReqHandler    planAstReqHandler;
    @Autowired
    PlanTranScreenRelDAO planTranScreenRelDAO;
    @Autowired
    PlanManager          planManager;
    @Autowired
    PlanRepo             planRepo;
    @Autowired
    PackRepository       packRepository;
    @Autowired
    TrCtrTaskRepo        trCtrTaskRepo;
    @Autowired
    AstPackManager       astPackManager;
    @Autowired
    private PlanTransferRuleManager planTransferRuleManager;
    @Autowired
    private OrgAcctDayComponent     orgAcctDayComponent;

    /**
     * 每天 0 0/5 * * * ?
     */
    @Value("${sync.quartz.plan.pack}")
    private String                  cron;

    @Override
    protected String getCronExpr() {
        return cron;
    }

    @Override
    protected boolean isRequestRecovery() {
        return true;
    }

    @Override
    public void doExecute() {
        // 转让会计日
        Date curDate = orgAcctDayComponent.getTrfAcctDay().getStandardDate();
    
        LogUtil.info(LOGGER, "[PlanPackJob]-[{0}]计划资产包任务start", curDate);
        List<String> orgCodes = astSyncTaskCtrRepo.getLastStatusFinishedOrgCodes(curDate, SyncTaskTypeEnum.AST_MOD);

        for (String orgCode : orgCodes) {
            LogUtil.info(LOGGER, "[PlanPackJob]-[{0}]执行管理机构[{1}]相关的计划资产包生成", curDate, orgCode);

            try {
                List<TrCtrTask> activeTrCtrTasks = trCtrTaskRepo
                        .getByManagerOrgCode(orgCode, curDate, TaskStatusEnum.INIT, AstPackTradeTypeEnum.AST_TRF_BIZ);
                List<Pack> activePacks = astPackManager.getByManagerOrgCode(orgCode, curDate, AstPackStatusEnum.INIT);

                if (CollectionUtils.isNotEmpty(activeTrCtrTasks) || CollectionUtils.isNotEmpty(activePacks)) {
                    LogUtil.warn(LOGGER, "[PlanPackJob]-[{0}]执行管理机构[{1}]资产包生成，有执行中的资产包[{2}]或者转让任务[{3}]", curDate,
                            orgCode, activePacks.size(), activeTrCtrTasks.size());
                    continue;
                }

                List<String> planNos = planTranScreenRelDAO.getPlanNosByManagerOrg(orgCode);

                List<Pack> packs = Lists.newArrayList();
                for (String planNo : planNos) {
                    // 获取计划全量信息
                    PlanBase planBase = planRepo.getPlanByNo(planNo);
                    // 放款池跳出不生成资产包
                    if (planBase.getPlanType() == PlanTypeEnum.SPV) {
                        LogUtil.debug(LOGGER, "[PlanPackJob]-跳过放款计划[{0}]", planBase.getPlanNo());
                        continue;
                    }
                    // 对应转让规则 及 筛选规则
                    PlanTransferRule transferRule = planTransferRuleManager.getByPlanNo(planNo);
                    VarChecker.checkNotNull(transferRule, "计划[{0}]创建资产包找不到对应转让规则", planBase.getPlanNo());
                    boolean trCheck = DateTimeUtil.between(curDate, transferRule.getStartTrDate(),
                            transferRule.getStopTrDate(), BetweenType.se);
                    if (trCheck) {
                        // 构建资产包
                        Pack pack = planAstReqHandler.execute(planBase, transferRule, curDate);
                        if (null != pack) {
                            packs.add(pack);
                        }
                    }
                }

                transactionTemplate.execute(new TransactionCallbackWithoutResult() {
                    @Override
                    protected void doInTransactionWithoutResult(TransactionStatus status) {
                        for (Pack pack : packs) {
                            LogUtil.info(LOGGER, "[PlanPackJob]-[{0}]生成资产包[{1}]", curDate, pack);
                            packRepository.create(pack);
                        }
                    }
                });
            } catch (Throwable e) {
                LogUtil.error(LOGGER, e, "[PlanPackJob]-[{0}]执行管理机构[{1}]相关的计划资产包异常", curDate, orgCode);
            }
        }

        LogUtil.info(LOGGER, "[PlanPackJob]-[{0}]计划资产包任务End", curDate);
    }
}