/**
 * abssqr.com Inc.
 * Copyright (c) 2017-2019 All Rights Reserved.
 */
package com.abssqr.plat.core.service.plan;

import com.abssqr.plat.common.facade.enums.CalcTypeEnum;
import com.abssqr.plat.common.facade.enums.EventCodeEnum;
import com.abssqr.plat.common.facade.enums.LayerTypeEnu;
import com.abssqr.plat.common.facade.enums.PlanAllocStatusEnum;
import com.abssqr.plat.common.facade.enums.PlanStatusEnum;
import com.abssqr.plat.common.facade.enums.TitleTypeEnum;
import com.abssqr.plat.common.facade.model.plan.FeeStructure;
import com.abssqr.plat.common.facade.model.plan.InvestStructure;
import com.abssqr.plat.common.facade.model.plan.PlanAcctTitle;
import com.abssqr.plat.common.facade.model.plan.PlanAlloc;
import com.abssqr.plat.common.facade.model.plan.PlanAllocItem;
import com.abssqr.plat.common.facade.model.plan.PlanBase;
import com.abssqr.plat.common.facade.model.plan.PlanCapitalOrder;
import com.abssqr.plat.common.facade.param.plan.alloc.PlanAllocHandleParam;
import com.abssqr.plat.common.facade.utils.CalcUtil;
import com.abssqr.plat.common.model.domain.loanTrans.TransInfoTemplate;
import com.abssqr.plat.common.model.domain.plan.PlanAcctTitleLog;
import com.abssqr.plat.common.model.domain.plan.PlanAcctTitleTransDetail;
import com.abssqr.plat.common.model.enums.IdTypeEnum;
import com.abssqr.plat.common.model.exception.AbssqrBizException;
import com.abssqr.plat.common.model.exception.AbssqrInvalidParamException;
import com.abssqr.plat.common.model.repo.plan.PlanAcctTitleLogRepo;
import com.abssqr.plat.common.model.repo.plan.PlanAcctTitleRepo;
import com.abssqr.plat.common.model.seq.SequenceUtil;
import com.abssqr.plat.core.service.system.TimeStampUtil;
import com.general.system.common.model.Money;
import com.general.system.common.util.DateTimeUtil;
import com.general.system.common.util.MessageUtil;
import com.general.system.common.util.VarChecker;
import com.general.system.tool.util.ToolUtil;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

/**
 * 计划结算科目管理
 *
 * @author zhenxuan.luo
 * @version com.abssqr.plat.core.service.plan: PlanAcctTitleManager.java, v 0.1 2019-06-01 16:01 zhenxuan.luo Exp $
 */
@Component
public class PlanAcctTitleManager {

    @Autowired
    private PlanAcctTitleRepo    planAcctTitleRepo;
    @Autowired
    private PlanAcctTitleLogRepo planAcctTitleLogRepo;
    @Autowired
    private PlanSettleManager    planSettleManager;
    @Autowired
    private PlanAllocManager     planAllocManager;

    /**
     * 记录计划结算科目
     *
     * @param transDetails
     */
    public void store(List<PlanAcctTitleTransDetail> transDetails) {

        List<PlanAcctTitleLog> logs = Lists.newArrayList();
        Map<String, PlanAcctTitle> planAcctTitleMap = Maps.newHashMap();
        for (PlanAcctTitleTransDetail transDetail : transDetails) {
            if (transDetail.isUpdate()) {
                PlanAcctTitle acctTitle = transDetail.getPlanAcctTitle();
                // 按时间顺序，覆盖map里的值，取最后一个acctTitle
                planAcctTitleMap.put(acctTitle.getTitleNo(), acctTitle);
                logs.add(transDetail.getActiveLog());
            }
        }

        for (PlanAcctTitle acctTitle : planAcctTitleMap.values()) {
            planAcctTitleRepo.updateAmt(acctTitle);
        }
        planAcctTitleLogRepo.insertBatch(logs);
    }

    /**
     * 根据计划-初始化科目信息
     *
     * @param planBase
     * @param effectDate       起息日
     * @param investStructures 投资结构
     * @param feeStructures    费用结构
     */
    public void initPlanAcctTitles(PlanBase planBase, Date effectDate, List<InvestStructure> investStructures,
                                   List<FeeStructure> feeStructures) {

        String planNo = planBase.getPlanNo();
        // 判断是否为准备中
        if (planBase.getStatus() == PlanStatusEnum.PREPARING){
            planAcctTitleRepo.cleanByPlanNo(planNo);
        }

        List<PlanAcctTitle> planAcctTitleList = planAcctTitleRepo.getByEntityNo(planNo);


        if (CollectionUtils.isEmpty(planAcctTitleList)) {

            planAcctTitleList = new ArrayList<>();
            // 1. 根据投资结构初始化科目
            for (InvestStructure investStructure : investStructures) {
                this.buildFromInvestStructure(planAcctTitleList, planNo, investStructure.getEffectDate(), investStructure);
            }

            // 2. 根据费用初始化科目
            for (FeeStructure feeStructure : feeStructures) {
                buildFromFee(planAcctTitleList, planNo, effectDate, feeStructure);
            }

            planAcctTitleRepo.insertBatch(planAcctTitleList);
        }else{
            List<String> refNoList = planAcctTitleList.stream().map(PlanAcctTitle::getRefNo).collect(Collectors.toList());
            List<PlanAcctTitle> newPlanAcctTitleList = Lists.newArrayList();
            // 1. 根据投资结构初始化科目
            for (InvestStructure investStructure : investStructures) {
                if(!refNoList.contains(investStructure.getLayerNo())) {
                    this.buildFromInvestStructure(newPlanAcctTitleList, planNo, investStructure.getEffectDate(), investStructure);
                }
            }

            // 2. 根据费用初始化科目
            for (FeeStructure feeStructure : feeStructures) {
                if(!refNoList.contains(feeStructure.getFeeNo())) {
                    buildFromFee(newPlanAcctTitleList, planNo, effectDate, feeStructure);
                }
            }
            if(CollectionUtils.isNotEmpty(newPlanAcctTitleList)) {
                planAcctTitleRepo.insertBatch(newPlanAcctTitleList);
            }
        }
    }

    /**
     * 根据投资结构构造计划科目
     *
     * @param planAcctTitleList
     * @param planNo
     * @param effectDate
     * @param investStructure
     */
    private void buildFromInvestStructure(final List<PlanAcctTitle> planAcctTitleList,
                                          String planNo, Date effectDate, InvestStructure investStructure) {

        //
        // 1. 投资结构科目
        PlanAcctTitle investTitle = buildBaseTitle(planNo);

        investTitle.setTitleName(MessageUtil.formatMsg("{0}-{1}",
            StringUtils.trim(investStructure.getLayerName()), TitleTypeEnum.INVESTED.getDesc()));
        investTitle.setRefNo(investStructure.getLayerNo());
        TitleTypeEnum investTitleTp = (investStructure.getLayerType() == LayerTypeEnu.PRIOR) ?
            TitleTypeEnum.PRIOR_INVESTED :
            TitleTypeEnum.SECND_INVESTED;
        investTitle.setTitleType(investTitleTp);
        investTitle.setBal(investStructure.getLayerAmt());
        investTitle.setRecpAmt(investStructure.getLayerAmt());
        investTitle.setInitAmt(investStructure.getLayerAmt());
        investTitle.setCalcType(CalcTypeEnum.FIX_AMT);
        investTitle.setEntryAmt(investStructure.getLayerAmt());
        investTitle.setTotalAmt(investStructure.getLayerAmt().getAmount());
        investTitle.setLastSettleDate(DateTimeUtil.getStandardDate(effectDate));
        investTitle.setEffectDate(effectDate);
        planAcctTitleList.add(investTitle);

        // 暂时不考虑浮动利率，如果是固定利率，构造投资人收益的科目，如果无票面利率，则不构造
        // 2. 投资机构对应的利息科目
        PlanAcctTitle intrTitle = buildBaseTitle(planNo);

        intrTitle.setTitleName(MessageUtil.formatMsg("{0}-{1}",
            StringUtils.trim(investStructure.getLayerName()), TitleTypeEnum.INVEST_PROFIT.getDesc()));

        intrTitle.setRefNo(investStructure.getLayerNo());
        intrTitle.setTitleType(TitleTypeEnum.INVEST_PROFIT);

        intrTitle.setRate(investStructure.getLayerIntr());

        CalcTypeEnum calcType;

        switch (investStructure.getIntrType()) {
            case FIXED:
                calcType = CalcTypeEnum.ANNU_RATE;
                break;
            case NOTICKET:
                calcType = CalcTypeEnum.NOTICKET;
                break;
            default:
                throw new AbssqrInvalidParamException(
                    MessageUtil.formatMsg("投资结构[{0}]配置了暂时不支持的利率类型", investStructure.getIntrType()));
        }
        intrTitle.setCalcType(calcType);
        intrTitle.setEffectDate(effectDate);

        planAcctTitleList.add(intrTitle);
    }

    /**
     * 根据费用结构构造计划科目
     *
     * @param planAcctTitleList
     * @param planNo
     * @param effectDate
     * @param feeStructure
     */
    private void buildFromFee(final List<PlanAcctTitle> planAcctTitleList,
                              String planNo, Date effectDate, FeeStructure feeStructure) {

        PlanSettleManager.RefInvestAmt refInvestAmt = planSettleManager.calcRefInvestment(planAcctTitleList,null);

        // 1. 投资结构科目
        PlanAcctTitle feeTitle = buildBaseTitle(planNo);

        feeTitle.setTitleName(feeStructure.getFeeName());
        feeTitle.setRefNo(feeStructure.getFeeNo());
        feeTitle.setTitleType(TitleTypeEnum.FEE);
        feeTitle.setScaleType(feeStructure.getFeeScaleType());
        feeTitle.setCalcType(feeStructure.getFeeCalcType());
        if (feeStructure.getFeeCalcType() == CalcTypeEnum.FIX_AMT) {
            feeTitle.setBal(feeStructure.getFixedFeeAmt());
            feeTitle.setRecpAmt(feeStructure.getFixedFeeAmt());
            feeTitle.setInitAmt(feeStructure.getFixedFeeAmt());
            feeTitle.setEntryAmt(feeStructure.getFixedFeeAmt());
            feeTitle.setTotalAmt(feeStructure.getFixedFeeAmt().getAmount());
            feeTitle.setLastSettleDate(DateTimeUtil.getStandardDate(effectDate));
            feeTitle.setScaleType(null);
        } else if(feeStructure.getFeeCalcType() == CalcTypeEnum.FIN_SER_FEE){
            feeTitle.setLastSettleDate(DateTimeUtil.getStandardDate(DateTimeUtil.addDays(effectDate,-1)));
            feeTitle.setFinSerFeeType(feeStructure.getFinSerFeeType());
        } else {
            feeTitle.setRate(feeStructure.getFeeRate());
            feeTitle.setUpperLimit(feeStructure.getPayLimitAmt());
        }
        if (feeStructure.getFeeCalcType() == CalcTypeEnum.FIX_RATE) {
            feeTitle.setRate(feeStructure.getFeeRate());
            Money baseAmt = refInvestAmt.getByScaleType(feeTitle.getScaleType());
            BigDecimal totalAmt = CalcUtil.getRealRate(feeTitle.getRate()).multiply(baseAmt.getAmount());
            feeTitle.setLastSettleDate(DateTimeUtil.getStandardDate(effectDate));
            feeTitle.entry(totalAmt);
        }
        // 默认费用起息日为项目起息日
        feeTitle.setEffectDate(effectDate);
        planAcctTitleList.add(feeTitle);
    }

    private PlanAcctTitle buildBaseTitle(String planNo) {
        PlanAcctTitle planAcctTitle = new PlanAcctTitle();

        planAcctTitle.setPlanNo(planNo);
        planAcctTitle.setTitleNo(SequenceUtil.genId(IdTypeEnum.ACCT_TITLE_CODE));
        return planAcctTitle;
    }

    /**
     * 变更分配计划状态 及 对应科目入账
     *
     * @param handleParam
     * @param order
     */
    public void allocTitleEntry(PlanAllocHandleParam handleParam, PlanCapitalOrder order) {
        PlanAlloc alloc = handleParam.getAlloc();
        VarChecker.checkEquals(alloc.getStatus(), PlanAllocStatusEnum.PENDING,
            "计划[{0}]当前分配计划[{1}]不能入账，状态为[{2}]",
                alloc.getPlanNo(), alloc.getAllocNo(), alloc.getStatus());

        VarChecker.checkEquals(alloc.getAllocAmt().getCent(), order.getCfrmAmt().getCent(),
            "计划[{0}]分配[{1}]金额不匹配-分配金额[{2}]-支付金额[{3}]",
                alloc.getPlanNo(), alloc.getAllocNo(), alloc.getAllocAmt(), order.getCfrmAmt());

        if (alloc.getStatus() != PlanAllocStatusEnum.PENDING) {
            throw new AbssqrBizException(MessageUtil.formatMsg("当前分配计划不能入账，状态为[{0}]", alloc.getStatus()));
        }
        List<PlanAllocItem> allocItems = alloc.getPlanAllocItemList();
        if (ToolUtil.isEmpty(allocItems)) {
            throw new AbssqrBizException(
                MessageUtil.formatMsg("分配计划兑付日[{0}-{1}]科目信息不存在", alloc.getPayDate(), alloc.getAllocNo()));
        }
        Map<String, List<PlanAllocItem>> itemsMap = allocItems.stream()
            .collect(Collectors.groupingBy(PlanAllocItem::getTitleNo));

        // 获取计划分配科目信息
        List<PlanAcctTitle> planAcctTitleList = planAcctTitleRepo.getByEntityNo(alloc.getPlanNo());
        if (ToolUtil.isEmpty(planAcctTitleList)) {
            throw new AbssqrBizException(MessageUtil.formatMsg("计划[{0}]科目信息不存在", alloc.getPlanNo()));
        }
        // 组装科目交易详情
        List<PlanAcctTitleTransDetail> transDetails = this.allocAssigned(
                order.getAccountDate(), order.getTransTime(), order.getOutTransNo(),
                alloc.getAllocNo(), itemsMap, planAcctTitleList);
        // 保存计划科目信息
        this.store(transDetails);

        alloc.setStatus(PlanAllocStatusEnum.ASSIGNED);
        // 变更分配计划为已分配
        planAllocManager.updateAllocStatus(alloc);
    }

    /**
     * 进行分配计算科目金额并累加
     *
     * @param accountDate
     * @param planAcctTitleList
     */
    private List<PlanAcctTitleTransDetail> allocAssigned(Date accountDate, Date transTime, String transNo,
                                                         String allocNo,
                                                         Map<String, List<PlanAllocItem>> itemsMap,
                                                         List<PlanAcctTitle> planAcctTitleList) {
        List<PlanAcctTitleTransDetail> result = Lists.newArrayList();
        for (PlanAcctTitle planAcctTitle : planAcctTitleList) {
            List<PlanAllocItem> items = itemsMap.get(planAcctTitle.getTitleNo());
            if (ToolUtil.isEmpty(items)) {
                continue;
            }
            // 科目分配兑付记流水
            TransInfoTemplate transInfoTemplate = new TransInfoTemplate();
            transInfoTemplate.setTransTime(transTime);
            transInfoTemplate.setAccountDate(DateTimeUtil.getStandardDate(accountDate));
            transInfoTemplate.setEventCode(EventCodeEnum.TITLE_ALLOC);
            transInfoTemplate.setOutBizNo(transNo);
            transInfoTemplate.setBizNo(allocNo);
            transInfoTemplate.setTimestamp(TimeStampUtil.getDefaultTimeStamp());
            PlanAcctTitleTransDetail transDetail = PlanAcctTitleTransDetail
                .createTxnDtl(planAcctTitle, transInfoTemplate, true);

            Money amt = PlanAlloc.calcPayAmt(items);
            Money taxAmt = PlanAlloc.calcPayVTaxAmt(items);
            // 记录科目金额变更
            transDetail.payAmt(amt, taxAmt);
            // 记录流水的变更后余额
            transDetail.processBeforeUpdate();
            result.add(transDetail);
        }
        return result;
    }
}