/**
 * abssqr.com Inc. 
 * Copyright (c) 2017-2019 All Rights Reserved. 
 */
package com.abssqr.plat.core.service.profit;

import com.abssqr.plat.common.facade.constants.AbssqrConstants;
import com.abssqr.plat.common.model.domain.ast.AstProfitUnit;
import com.abssqr.plat.common.model.utils.DecimalCalcUtil;
import com.abssqr.plat.common.util.utils.BigDecimalUtil;
import com.general.system.common.model.Money;
import com.general.system.common.util.MoneyUtil;
import com.general.system.common.util.VarChecker;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import org.springframework.stereotype.Component;

import java.math.BigDecimal;
import java.util.Collections;
import java.util.List;
import java.util.Map;

/**
 * 收益分润计算器
 * @author zhenxuan.luo
 * @version com.abssqr.plat.core.service.profit: ProfitSplitCalculator.java, v 0.1 2019-11-17 下午16:37 zhenxuan.luo Exp $ 
 */
@Component
public class ProfitSplitCalculator {
    
    /**
     * 按金额进行收益重组分配
     * @param fieldAmt 待分配金额
     * @param entriedAmt 已分配金额
     * @return
     */
    public <T> void splitDebtMoney(Money fieldAmt, Map<T, AstProfitUnit> entriedAmt) {
        if (fieldAmt.isZero()) {
            return;
        }
        VarChecker.checkArgument(entriedAmt != null && entriedAmt.size() > 0,
                "按金额分配必须传入对应的分配结果项");
        Map<T, AstProfitUnit> ret = entriedAmt;
        
        /**
         * 计算分配比例
         */
        Map<T, BigDecimal> ratios = Maps.newHashMap();
        DecimalCalcUtil.calcSplitFactor(ratios, filterAmtBaseDist(entriedAmt), true);
        
        /**
         * 计算实数金额
         */
        BigDecimal fieldDec = fieldAmt.getAmount();
        Money remain = fieldAmt.clone();
        Money checkEntry = new Money();
        BigDecimal checkEntryDec = BigDecimal.ZERO;
        BigDecimal remainDec = BigDecimalUtil.clone(fieldDec);
        
        for (T key : ret.keySet()) {
            AstProfitUnit profitUnit = ret.get(key);
            
            // 计算入账实数
            BigDecimal ratio = ratios.get(key);
            BigDecimal curEntryDec = fieldDec.multiply(ratio);
            curEntryDec = BigDecimalUtil.min(curEntryDec, remainDec);
            remainDec = remainDec.subtract(curEntryDec);
            
            // 计算入账金额
            Money curEntryAmt = new Money(curEntryDec, AbssqrConstants.DECIMAL_ROUNDING_MODE);
            curEntryAmt = MoneyUtil.min(curEntryAmt, remain);
            remain = remain.subtract(curEntryAmt);
            
            // 入账
            checkEntry.addTo(curEntryAmt);
            checkEntryDec = checkEntryDec.add(curEntryDec);
            profitUnit.trans(curEntryAmt, curEntryDec);
            
            if (remain.isZero()) {
                checkEntryDec = checkEntryDec.add(remainDec);
                profitUnit.trans(remain, remainDec);
                break;
            }
        }
        
        if (remain.isGreaterThanZero()) {
            
            List<AstProfitUnit> profitUnitList = Lists.newArrayList(ret.values());
            
            Collections.sort(profitUnitList, (o1, o2) -> {
                BigDecimal diffAmt1 = o1.getEntryDec().subtract(o1.getEntryAmt().getAmount());
                BigDecimal diffAmt2 = o2.getEntryDec().subtract(o2.getEntryAmt().getAmount());
                return diffAmt2.compareTo(diffAmt1);
            });
            
            for (AstProfitUnit profitUnit : profitUnitList) {
                BigDecimal diffAmt = profitUnit.getEntryDec().subtract(profitUnit.getEntryAmt().getAmount());
                Money entryAmt = new Money();
                BigDecimal entryDec = BigDecimal.ZERO;
                
                if (BigDecimalUtil.isNotNegative(diffAmt)) {
                    entryAmt = new Money(diffAmt);
                    if (entryAmt.isZero()) {
                        entryAmt = Money.createWithCent(1L);
                    }
                    
                    entryAmt = MoneyUtil.min(entryAmt, remain);
                } else {
                    entryAmt = Money.createWithCent(1L);
                    entryAmt = MoneyUtil.min(entryAmt, remain);
                }
                
                Money newEntryAmt = entryAmt.add(profitUnit.getEntryAmt());
                if (BigDecimalUtil.isGT(newEntryAmt.getAmount(), profitUnit.getEntryDec())) {
                    entryDec = newEntryAmt.getAmount().subtract(profitUnit.getEntryDec());
                    entryDec = BigDecimalUtil.min(entryDec, remainDec);
                }
                
                checkEntry.addTo(entryAmt);
                checkEntryDec = checkEntryDec.add(entryDec);
                
                profitUnit.trans(entryAmt, entryDec);
                remain = remain.subtract(entryAmt);
                remainDec = remainDec.subtract(entryDec);
                
                if (remain.isZero()) {
                    checkEntryDec = checkEntryDec.add(remainDec);
                    profitUnit.trans(remain, remainDec);
                    break;
                }
            }
        }
        
        VarChecker.checkEquals(fieldAmt, checkEntry,
                "[splitDebtMoney]待分配金额[{0}]和实际分配金额汇总[{1}]不相等", fieldAmt, checkEntry);
        VarChecker.checkArgument(BigDecimalUtil.isEQ(fieldAmt.getAmount(), checkEntryDec),
                "[splitDebtMoney]待分配实数金额[{0}]实际分配实数金额汇总[{1}]不相等", fieldAmt, checkEntryDec);
    }
    
    /**
     * 按收益单元分配
     * @param transProfitUnit
     * @param entriedAmt
     */
    public void splitProfitUnit(AstProfitUnit transProfitUnit,
                                Map<String, AstProfitUnit> entriedAmt) {
        VarChecker.checkArgument(entriedAmt != null && entriedAmt.size() > 0,
                "按收益单元分配必须传入对应的分配结果项");
        Map<String, AstProfitUnit> ret = entriedAmt;
        
        /**
         * 计算分配比例
         */
        Map<String, BigDecimal> ratios = Maps.newHashMap();
        DecimalCalcUtil.calcSplitFactor(ratios, filterAmtBaseDist(entriedAmt), true);
        
        /**
         * 计算实数金额
         */
        // 始终使用总入账DEC去计算分布，不然会导致始终将精度的进舍放在一个单元上
        BigDecimal remainDec = BigDecimalUtil.clone(transProfitUnit.getTransEntryDec());
        Money remainAmt = transProfitUnit.getTransEntryAmt().clone();
        BigDecimal remainRatio = BigDecimal.ONE;
        
        BigDecimal entryDec = transProfitUnit.getTransEntryDec();
        
        BigDecimal checkEntryDec = BigDecimal.ZERO;
        Money checkEntryAmt = new Money();
        BigDecimal checkTotalEntryDec = BigDecimal.ZERO;
        Money checkTotalEntryAmt = new Money();
        
        for (String key : ret.keySet()) {
            AstProfitUnit profitUnit = ret.get(key);
            
            if (profitUnit.getBase().isZero()) {
                checkTotalEntryAmt.addTo(profitUnit.getEntryAmt());
                checkTotalEntryDec = checkTotalEntryDec.add(profitUnit.getEntryDec());
                continue;
            }
            // 计算入账实数
            // 1. 确定比例
            BigDecimal ratio = ratios.get(key);
            ratio = BigDecimalUtil.min(ratio, remainRatio);
            remainRatio = remainRatio.subtract(ratio);
            
            // 2. 确定入账实数
            BigDecimal curEntryDec = entryDec.multiply(ratio);
            curEntryDec = BigDecimalUtil.min(curEntryDec, remainDec);
            remainDec = remainDec.subtract(curEntryDec);
            
            // 计算总入账金额
            BigDecimal totalEntryDec = profitUnit.getEntryDec().add(curEntryDec);
            
            // 3. 计算入账金额
            Money totalEntryAmt = new Money(totalEntryDec, AbssqrConstants.DECIMAL_ROUNDING_MODE);
            
            Money curEntryAmt = totalEntryAmt.subtract(profitUnit.getEntryAmt());
            curEntryAmt = MoneyUtil.max(curEntryAmt, new Money());
            curEntryAmt = MoneyUtil.min(curEntryAmt, remainAmt);
            remainAmt = remainAmt.subtract(curEntryAmt);
            
            // 入账
            checkEntryDec = checkEntryDec.add(curEntryDec);
            checkEntryAmt.addTo(curEntryAmt);
            profitUnit.trans(curEntryAmt, curEntryDec);
            
            checkTotalEntryAmt.addTo(profitUnit.getEntryAmt());
            checkTotalEntryDec = checkTotalEntryDec.add(profitUnit.getEntryDec());
        }
        
        // DEC不会有多余，因为上面保证了所有分配完毕
        if (remainAmt.isGreaterThanZero()) {
            // 获取入账实数和入账金额差异最大的项
            String maxDiffIndex = getMaxEntryDiffUnit(ret);
            checkEntryAmt.addTo(remainAmt);
            checkTotalEntryAmt.addTo(remainAmt);
            
            checkEntryDec = checkEntryDec.add(remainDec);
            checkTotalEntryDec = checkTotalEntryDec.add(remainDec);
            // 将剩余尾数金额放入对应的收益单元
            ret.get(maxDiffIndex).trans(remainAmt, remainDec);
        }
        
        VarChecker.checkEquals(transProfitUnit.getTransEntryAmt(), checkEntryAmt,
                "[splitProfitUnit]待分配金额发生额[{0}]和实际分配金额汇总[{1}]不相等", transProfitUnit.getTransEntryAmt(), checkEntryAmt);
        VarChecker.checkArgument(BigDecimalUtil.isEQ(transProfitUnit.getTransEntryDec(), checkEntryDec),
                "[splitProfitUnit]待分配实数发生额[{0}]和实际分配实数汇总[{1}]不相等", transProfitUnit.getTransEntryDec(), checkEntryDec);
        VarChecker.checkEquals(transProfitUnit.getEntryAmt(), checkTotalEntryAmt,
                "[splitProfitUnit]待分配金额总额[{0}]和实际分配金额汇总[{1}]不相等", transProfitUnit.getEntryAmt(), checkTotalEntryAmt);
        VarChecker.checkArgument(BigDecimalUtil.isEQ(transProfitUnit.getEntryDec(), checkTotalEntryDec),
                "[splitProfitUnit]待分配实数总额[{0}]和实际分配实数汇总[{1}]不相等", transProfitUnit.getEntryDec(), checkTotalEntryDec);
    }
    
    /**
     * 获取入账实数和入账金额差异最大的项（被舍弃的金额最多的一个）
     * @param profitUnitMap
     * @param <T>
     * @return
     */
    private <T> T getMaxEntryDiffUnit(Map<T, AstProfitUnit> profitUnitMap) {
        BigDecimal maxDiffAmt = null;
        T maxItemIdx = null;
        
        for (T key : profitUnitMap.keySet()) {
            AstProfitUnit profitUnit = profitUnitMap.get(key);
            if (profitUnit.getBase().isZero()) {
                continue;
            }
            BigDecimal diffAmt = profitUnit.getEntryDec().subtract(profitUnit.getEntryAmt().getAmount());
            if (maxItemIdx == null || BigDecimalUtil.isGT(diffAmt, maxDiffAmt)) {
                maxItemIdx = key;
                maxDiffAmt = diffAmt;
            }
        }
        
        return maxItemIdx;
    }
    
    /**
     * 构造
     * @param source
     * @param <T>
     * @return
     */
    private <T> Map<T, AstProfitUnit> copyStructure(Map<T, AstProfitUnit> source) {
        Map<T, AstProfitUnit> ret = Maps.newHashMap();
        for (T key : source.keySet()) {
            AstProfitUnit sourceUnit = source.get(key);
            ret.put(key, new AstProfitUnit(sourceUnit.getField()));
            ret.get(key).append(sourceUnit);
        }
        
        return ret;
    }
    
    /**
     * 抽取分配计算的基数
     *
     * @param profitUnitMap
     * @param <T>
     * @return
     */
    private <T> Map<T, BigDecimal> filterAmtBaseDist(Map<T, AstProfitUnit> profitUnitMap) {
        Map<T, BigDecimal> res = Maps.newHashMap();
        
        for (T key : profitUnitMap.keySet()) {
            AstProfitUnit measure = profitUnitMap.get(key);
            res.put(key, measure.getBase().getAmount());
        }
        
        return res;
    }
}