/**
 * abssqr.com Inc. 
 * Copyright (c) 2017-2020 All Rights Reserved. 
 */
package com.abssqr.plat.core.service.rule;

import com.abssqr.plat.common.dal.mysql.auto.resultmap.AstPoolItem;
import com.abssqr.plat.common.facade.enums.ScreenTypeEnum;
import com.abssqr.plat.common.facade.model.plan.PlanTransferRule;
import com.abssqr.plat.common.model.domain.ast.AstConditions;
import com.abssqr.plat.common.model.domain.ast.Loan;
import com.abssqr.plat.common.model.domain.ast.LoanAstGroup;
import com.abssqr.plat.common.model.domain.ast.Pack;
import com.abssqr.plat.common.model.domain.task.TrTask;
import com.abssqr.plat.common.model.utils.TextUtil;
import com.abssqr.plat.core.service.astscreen.AstScreenRuleChecker;
import com.general.system.common.util.DateTimeUtil;
import com.general.system.common.util.LogUtil;
import com.general.system.common.util.MessageUtil;
import com.general.system.common.util.MoneyUtil;
import com.google.common.collect.Maps;
import org.apache.commons.collections.CollectionUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.util.Date;
import java.util.Map;

/**
 * 资产筛选匹配组件
 * @author hanfei
 * @version $Id: AstScreenRuleComponent.java, v 0.1 2020-02-07 6:01 PM hanfei Exp $
 */
@Component
public class AstScreenRuleComponent {
    
    private final static Logger LOGGER = LoggerFactory.getLogger(AstScreenRuleComponent.class);
    
    @Autowired
    private AstScreenRuleChecker astScreenRuleChecker;
    
    /**
     * 资产转让规则匹配及筛选规则匹配
     * @param acctDate 会计日
     * @param pack 资产包
     * @param assemble 组装AstPoolItem扩展方法
     * @return
     */
    public boolean astMatchTransferRule(Date acctDate, Pack pack, AstPoolItemAssemble assemble){
        return astScreenRuleChecker.matchWithTransferPack(acctDate, assemble.getAstPoolItem(), pack, ScreenTypeEnum.AST_TRF);
    }
    
    /**
     * 回购资产筛选规则匹配
     * @param acctDate
     * @param pack
     * @param assemble
     * @return
     */
    public boolean astBuybackScreen(Date acctDate, Pack pack, AstPoolItemAssemble assemble){
        AstPoolItem astPoolItem = assemble.getAstPoolItem();
        Map<String, String> context = TextUtil.convertTxt2Map(astPoolItem.getContext());
        return astScreenRuleChecker.astScreenRuleMatch(acctDate, assemble.getAstPoolItem(), context,
                pack, pack.getTransferRule().getAstScreenRuleList(), ScreenTypeEnum.BUY_BACK);
    }
    
    /**
     * 校验转让规则
     *
     * @param loan
     * @param loanAstGroup
     * @param trTask
     * @param pack
     * @return
     */
    public boolean checkTransfer(Loan loan, LoanAstGroup loanAstGroup, TrTask trTask,
                                 Pack pack, PlanTransferRule planTransferRule) {
        // 1、基础条件校验
        if (!this.preCheck( loanAstGroup, trTask, loan.getLastUpdateDate())){
            return false;
        }
        // 2. 转让规则与筛选规则匹配
        if (!this.astMatchTransferRule(trTask.getTfrDt(), pack, () -> {
            AstPoolItem astPoolItem = buildAstPoolItem(loan, loanAstGroup, trTask);
            return astPoolItem;
        })) {
            LogUtil.warn(LOGGER, "[AstTransferIgnore]资产转让失败[{0}]，不符合转让条件", trTask);
            return false;
        }
    
        // 3. 是否支持切片校验 非切片情况必须转让整笔资产
        if (!planTransferRule.getSlice()) {
            if (!loanAstGroup.isAllAst()) {
                LogUtil.warn(LOGGER, "[AstTransferIgnore]资产转让未包含所有资产，且[{1}]计划不能接受部分转让[{0}]", trTask,
                        planTransferRule.getPlanNo());
                return false;
            }
        }
        return true;
    }
    
    /**
     * 回购交易筛选规则校验
     * @param loan
     * @param loanAstGroup
     * @param trTask
     * @param pack
     * @return
     */
    public boolean checkBuyback(Loan loan, LoanAstGroup loanAstGroup, TrTask trTask, Pack pack){
        // 1、基础条件校验
        if (!this.preCheck( loanAstGroup, trTask, loan.getLastUpdateDate())){
            return false;
        }
        // 2. 筛选规则匹配
        if (!this.astBuybackScreen(trTask.getTfrDt(), pack, () -> {
            AstPoolItem astPoolItem = buildAstPoolItem(loan, loanAstGroup, trTask);
            return astPoolItem;
        })) {
            LogUtil.warn(LOGGER, "[AstTransferIgnore]资产转让失败[{0}]，不符合转让条件", trTask);
            return false;
        }
        return true;
    }
    
    
    
    /**
     * 基础前置校验
     *
     * @param loanAstGroup
     * @param trTask
     * @return
     */
    private boolean preCheck(LoanAstGroup loanAstGroup, TrTask trTask, Date loanLastUpdateDate) {
        // 计算日期
        if (DateTimeUtil.calcDayDiff(trTask.getTfrDt(), loanLastUpdateDate) != 0) {
            LOGGER.warn(MessageUtil.formatMsg("[AstTransferIgnore]资产最近变更会计日期[{1}]小于转让日，不执行[{0}]", trTask,
                    loanLastUpdateDate));
            return false;
        }
        // 1. 判断是否可以执行转让
        if (CollectionUtils.isEmpty(loanAstGroup.getLoanAsts())) {
            LOGGER.warn(MessageUtil.formatMsg("[AstTransferIgnore]捞不到可转资产或已经被转让[{0}]", trTask));
            return false;
        }
        // 4. 转让金额校验
        if (MoneyUtil.compare(loanAstGroup.getPrinBal(), trTask.getTfrPrinBal()) > 0) {
            LogUtil.warn(LOGGER, "[AstTransferIgnore]资产转让资产本金[{1}]大于待转本金，可能导致不足转让兑付[{0}]", trTask,
                    loanAstGroup.getPrinBal());
            return false;
        }
        return true;
    }
    
    /**
     * 构建资产条目
     * @param loan
     * @param loanAstGroup
     * @param trTask
     * @return
     */
    private AstPoolItem buildAstPoolItem(Loan loan, LoanAstGroup loanAstGroup, TrTask trTask) {
        AstPoolItem astPoolItem = new AstPoolItem();
        astPoolItem.setStatus(loan.getStatus().getCode());
        astPoolItem.setAstNo(loan.getAstNo());
        astPoolItem.setProdNo(loan.getProdNo());
        astPoolItem.setTermNo(loanAstGroup.getMaxTerm());
        astPoolItem.setStartDate(loanAstGroup.getFirstInstal().getAstPeriod().getStartDate());
        astPoolItem.setEndDate(loanAstGroup.getLastInstal().getAstPeriod().getEndDate());
        astPoolItem.setRemainDays((int) DateTimeUtil.calcDayDiff(astPoolItem.getEndDate(), trTask.getTfrDt()));
        astPoolItem.setPrinBal(loanAstGroup.getPrinBal());
        //astPoolItem.setIntrBal(loan.getBal().getAllIntr());
        //astPoolItem.setPrinPenalBal(loan.getBal().getByFieldType(DebtMoneyFieldEnum.PRIN_ODUE_PNY));
        //astPoolItem.setIntrPenalBal(loan.getBal().getByFieldType(DebtMoneyFieldEnum.INTR_ODUE_PNY));
        astPoolItem.setManagerOrgCode(loan.getAstOrg().getManagerOrgCode());
        astPoolItem.setLenderOrgCode(loan.getAstOrg().getAstOwnership().getLenderOrgCode());
        astPoolItem.setTermCnt(loan.getTermCnt());
        astPoolItem.setRiskCl(loan.getLoanRisk().getRiskClass().getCode());
        Map<String, String> context = Maps.newHashMap();
        context.put(AstConditions.CERT_NO, loan.getContract().getLoanCustInfo().getCertNo());
        context.put(AstConditions.CERT_TYPE, loan.getContract().getLoanCustInfo().getCertType().getCode());
        context.put(AstConditions.INTER_RATE, loan.getContract().getLoanProperty().getRate().toString());
        context.put(AstConditions.RATE_TYPE, loan.getContract().getLoanProperty().getRateType().getCode());
        astPoolItem.setContext(TextUtil.convertMap2Txt(context));
        
        return astPoolItem;
    }
    
    public interface AstPoolItemAssemble{
        AstPoolItem getAstPoolItem();
    }
}
