/**
 * abssqr.com Inc.
 * Copyright (c) 2017-2020 All Rights Reserved.
 */
package com.abssqr.plat.core.service.util;

import com.abssqr.plat.common.facade.utils.CalcUtil;
import com.abssqr.plat.common.model.domain.rule.LoanAstBalAndPrice;
import com.abssqr.plat.common.model.domain.task.TrTask;
import com.general.system.common.model.Money;
import com.general.system.common.util.LogUtil;
import com.general.system.common.util.MoneyUtil;
import com.general.system.tool.util.ToolUtil;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.math.BigDecimal;

/**
 * 计价金额与发生校验
 *
 * @author hanfei
 * @version $Id: AstPriceCalcAmtUtil.java, v 0.1 2020-02-14 12:45 AM hanfei Exp $
 */
public class AstPriceCalcAmtUtil {

    private final static Logger logger = LoggerFactory.getLogger(AstPriceCalcAmtUtil.class);

    /**
     * 校验定价金额 与 计价金额
     *
     * @param balAndPrice
     * @param task
     * @return
     */
    public static boolean checkGreaterThan(LoanAstBalAndPrice balAndPrice, TrTask task) {
        if (null == balAndPrice) {
            LogUtil.info(logger, "资产回购交易定价金额为空astNo[{0}]",
                    task.getAstNo());
            return true;
        }
        // 比较本金余额
        if (MoneyUtil.compare(balAndPrice.getBal().getAllPrin(), task.getTfrPrinBal()) > 0) {
            LogUtil.info(logger, "资产回购交易本金大于计价本金astNo[{0}]out[{1}]=>amt[{2}]price[{3}]",
                    task.getAstNo(), task.getTfrOut(), balAndPrice.getBal().getAllPrin(), task.getTfrPrinBal());
            return true;
        }
        // 比较利息余额
        if (MoneyUtil.compare(balAndPrice.getBal().getAllIntr(), task.getTfrIntrBal()) > 0) {
            LogUtil.info(logger, "资产回购交易利息大于计价利息astNo[{0}]out[{1}]=>amt[{2}]price[{3}]",
                    task.getAstNo(), task.getTfrOut(), balAndPrice.getBal().getAllIntr(), task.getTfrIntrBal());
            return true;
        }
        // 比较费用余额
        Money penalFeeVal = balAndPrice.getBal().getAllIntAndFee().subtract(balAndPrice.getBal().getAllIntr());
        if (MoneyUtil.compare(penalFeeVal, task.getTfrPenalFeeBal()) > 0) {
            LogUtil.info(logger, "资产回购交易费用大于计价费用astNo[{0}]out[{1}]=>amt[{2}]price[{3}]",
                    task.getAstNo(), task.getTfrOut(), penalFeeVal, task.getTfrPenalFeeBal());
            return true;
        }
    
        // 比较定价总额
        if (MoneyUtil.compare(balAndPrice.getPrice().getSum(), task.getTfrAmt()) > 0) {
            LogUtil.info(logger, "资产回购交易定价金额大于计价定价astNo[{0}]out[{1}]=>amt[{2}]price[{3}]",
                    task.getAstNo(), task.getTfrOut(), balAndPrice.getPrice().getSum(), task.getTfrAmt());
            return true;
        }
        return false;
    }


    /**
     * 按比例计算金额
     *
     * @param amt  计算基数金额
     * @param rate 比例 如 0.95
     * @return
     */
    public static Money calcPrice(Money amt, BigDecimal rate) {
        // 为空和1时为原价
        if (ToolUtil.isEmpty(rate) ||
                BigDecimal.ONE.compareTo(rate) == 0) {
            return amt;
        }
        // 按比例计算
        return CalcUtil.calcMoneyRatio(amt, rate);
    }
}
